/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001-2005  by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@gmx.de
 *    Borg Enders  borg@borgsoft.de
 *
 ********************************************************************/

#ifdef NETWORK

#include "constants.h"

#include "server.h"

#include "../ui/ui.h"

typedef struct
{
  GTcpSocket* socket;
  gchar* name;
  guint watch_flags;
  gint port;
  guint watch;
  gchar buffer[1024];
  guint n;

} ClientState;


static void clientstate_delete (ClientState* state);

static void async_accept (GTcpSocket* server_socket, 
			  GTcpSocket* client_socket,
			  gpointer data);
static gboolean async_client_iofunc (GIOChannel* client, 
				     GIOCondition condition, gpointer data);



namespace Network {

  /**
   ** ...
   **
   ** @param	-
   **
   ** @return	-
   **
   ** @version	0.6.9
   **
   ** @author	Borg Enders
   ** @author	Diether Knof
   **/
  Server::Server() :
      local_addr(),
      port(DEFAULT_PORT),
      iochannels_clients()
  {
    GTcpSocket* server = gnet_tcp_socket_server_new_with_port(this->port);
    if (server == NULL) {
      cerr << "Could not create server on port " << port << "." << endl;
      exit (EXIT_FAILURE);
    }
    this->local_addr = gnet_tcp_socket_get_local_inetaddr(server);

    /* Wait asyncy for incoming clients */
    gnet_tcp_socket_server_accept_async (server, async_accept, NULL);

    return ;
  } // Server::Server()

  Server::~Server()
  {
  }

} // namespace Network





static void
clientstate_delete (ClientState* state)
{
  g_source_remove (state->watch);
  gnet_tcp_socket_delete (state->socket);
  g_free (state->name);
  g_free (state);
}

static void
async_accept (GTcpSocket* server, GTcpSocket* client, gpointer data)
{
  if (client)
  {
    GInetAddr*  addr;
    gchar*      name;
    gint	  port;
    GIOChannel* client_iochannel;
    ClientState* client_state;

    /* Print the address */
    addr = gnet_tcp_socket_get_local_inetaddr(client);
    g_assert (addr);
    name = gnet_inetaddr_get_canonical_name (addr);
    g_assert (name);
    port = gnet_inetaddr_get_port (addr);
    gnet_inetaddr_delete (addr);

    client_iochannel = gnet_tcp_socket_get_io_channel (client);
    g_assert (client_iochannel != NULL);

    client_state = g_new0(ClientState, 1);
    client_state->socket = client;
    client_state->name = name;
    client_state->port = port;
    client_state->watch_flags = G_IO_IN | G_IO_ERR | G_IO_HUP | G_IO_NVAL;
    client_state->watch = 
      g_io_add_watch (client_iochannel, static_cast<GIOCondition>(client_state->watch_flags),
		      async_client_iofunc, client_state);

    g_print ("Accepted connection from %s:%d\n", name, port);
  }
  else
  {
    fprintf (stderr, "Server error\n");
    exit (EXIT_FAILURE);
  }
}


/*

   Client IO callback.  Called for errors, input, or output.  When
   there is input, we reset the watch with the output flag (if we
   haven't already).  When there is no more data, the watch is reset
   without the output flag.

*/
gboolean
async_client_iofunc (GIOChannel* iochannel, GIOCondition condition, 
		     gpointer data)
{
  ClientState* client_state = (ClientState*) data;
  gboolean rv = TRUE;

  g_assert (client_state != NULL);


  /* Check for socket error */
  if (condition & (G_IO_ERR | G_IO_HUP | G_IO_NVAL))
  {
    fprintf (stderr, "Client socket error (%s:%d)\n", 
	     client_state->name, client_state->port);
    goto error;
  }

  /* Check for data to be read (or if the socket was closed) */
  if (condition & G_IO_IN)
  {
    GIOError error;
    guint bytes_read;

    /* Read the data into our buffer */
    error = 
      g_io_channel_read (iochannel, 
			 &client_state->buffer[client_state->n], 
			 sizeof(client_state->buffer) - client_state->n, 
			 &bytes_read);

    string const text(&client_state->buffer[client_state->n], bytes_read);
    cout << "Network text: " << text;
    ::ui->chat("network", text);


    /* Check for socket error */
    if (error != G_IO_ERROR_NONE)
    {
      fprintf (stderr, "Client read error (%s:%d): %d\n", 
	       client_state->name, client_state->port, error);
      goto error;
    }

    /* Check if we read 0 bytes, which means the connection is
       closed */
    else if (bytes_read == 0)
    {
      g_print ("Connection from %s:%d closed\n", 
	       client_state->name, client_state->port);
      goto error;
    }

    /* Otherwise, we read something */
    else
    {
      guint old_watch_flags;

      g_assert (bytes_read > 0);

      old_watch_flags = client_state->watch_flags;
      client_state->n += bytes_read;

      /* Add an add watch */
      client_state->watch_flags |= G_IO_OUT;

      /* Remove the IN watch if the buffer is full */
      if (client_state->n == sizeof(client_state->buffer))
	client_state->watch_flags &= ~G_IO_IN;

      /* Update watch flags if they changed */
      if (old_watch_flags != client_state->watch_flags)
      {
	g_source_remove (client_state->watch);
	client_state->watch = 
	  g_io_add_watch(iochannel, static_cast<GIOCondition>(client_state->watch_flags),
			 async_client_iofunc, client_state);
	rv = FALSE;
      }
    }

  }

  if (condition & G_IO_OUT)
  {
    GIOError error;
    guint bytes_written;

    /* Write the data out */
    error = g_io_channel_write(iochannel, client_state->buffer, 
			       client_state->n, &bytes_written);

    if (error != G_IO_ERROR_NONE) 
    {
      fprintf (stderr, "Client write error (%s:%d): %d\n", 
	       client_state->name, client_state->port, error);
      goto error;
    }

    else if (bytes_written > 0)
    {
      guint old_watch_flags;

      /* Move the memory down some (you wouldn't want to do this
	 in a performance server because it's slow!) */
      client_state->n -= bytes_written;
      g_memmove(client_state->buffer, 
		&client_state->buffer[bytes_written],
		client_state->n);

      old_watch_flags = client_state->watch_flags;

      /* Remove OUT watch if done */
      if (client_state->n == 0)
	client_state->watch_flags &= ~G_IO_OUT;

      /* Add IN watch */
      client_state->watch_flags |= G_IO_IN;

      /* Update watch flags if they changed */
      if (old_watch_flags != client_state->watch_flags)
      {
	g_source_remove (client_state->watch);
	client_state->watch = 
	  g_io_add_watch(iochannel, static_cast<GIOCondition>(client_state->watch_flags),
			 async_client_iofunc, client_state);
	rv = FALSE;
      }
    }
  }

  return rv;

error:
  clientstate_delete (client_state);
  return FALSE;
}

#endif // #ifdef NETWORK
